'''
Created on Jan 19, 2020

@author: ballance
'''

import argparse
import os
import sys

from ivpm.packages_info import PackagesInfo
from ivpm.proj_info import ProjInfo
from .cmds.cmd_activate import CmdActivate
from .cmds.cmd_build import CmdBuild
from .cmds.cmd_init import CmdInit
from .cmds.cmd_update import CmdUpdate
from .cmds.cmd_git_status import CmdGitStatus
from .cmds.cmd_git_update import CmdGitUpdate
from .cmds.cmd_share import CmdShare
from .cmds.cmd_snapshot import CmdSnapshot
from .cmds.cmd_pkg_info import CmdPkgInfo


def get_parser(parser_ext = None):
    """Create the argument parser"""
    parser = argparse.ArgumentParser(prog="ivpm")
    
    subparser = parser.add_subparsers()
    subparser.required = True
    subparser.dest = 'command'

    activate_cmd = subparser.add_parser("activate",
        help="Starts a new shell that contains the activated python virtual environment")
    activate_cmd.add_argument("-c",
        help="When specified, executes the specified string")
    activate_cmd.add_argument("-p", "--project-dir", dest="project_dir",
        help="Specifies the project directory to use (default: cwd)")
    activate_cmd.add_argument("args", nargs='*')
    activate_cmd.set_defaults(func=CmdActivate())

    build_cmd = subparser.add_parser("build",
        help="Build all sub-projects with an IVPM-supported build infrastructure (Python)")
    build_cmd.add_argument("-d", "--dep-set", dest="dep_set", 
        help="Uses dependencies from specified dep-set instead of 'default-dev'")
    build_cmd.add_argument("-g", "--debug", 
        action="store_true",
        help="Enables debug for native extensions")
    build_cmd.set_defaults(func=CmdBuild())

    pkginfo_cmd = subparser.add_parser("pkg-info",
        help="Collect paths/files for a listed set of packages")
    pkginfo_cmd.add_argument("type", 
            choices=("paths", "libdirs", "libs", "flags"),
            help="Specifies what info to query")
    pkginfo_cmd.add_argument("-k", "--kind",
            help="Specifies qualifiers on the type of info to query")
    pkginfo_cmd.add_argument("pkgs", nargs="+")
    pkginfo_cmd.set_defaults(func=CmdPkgInfo())

    share_cmd = subparser.add_parser("share",
        help="Returns the 'share' directory, which includes cmake files, etc")
    share_cmd.add_argument("path", nargs=argparse.REMAINDER)
    share_cmd.set_defaults(func=CmdShare())

    update_cmd = subparser.add_parser("update",
        help="Fetches packages specified in ivpm.yaml that have not already been loaded")
    update_cmd.set_defaults(func=CmdUpdate())
    update_cmd.add_argument("-p", "--project-dir", dest="project_dir",
        help="Specifies the project directory to use (default: cwd)")
    update_cmd.add_argument("-d", "--dep-set", dest="dep_set", 
        help="Uses dependencies from specified dep-set instead of 'default-dev'")
    update_cmd.add_argument("-a", "--anonymous-git", dest="anonymous", 
        action="store_true",
        help="Clones git repositories in 'anonymous' mode")
#    update_cmd.add_argument("-r", "--requirements", dest="requirements")
    
    init_cmd = subparser.add_parser("init",
        help="Creates an initial ivpm.yaml file")
    init_cmd.set_defaults(func=CmdInit())
    init_cmd.add_argument("-v", "--version", default="0.0.1")
    init_cmd.add_argument("-f", "--force", default=False, action='store_const', const=True)
    init_cmd.add_argument("name")
    
    git_status_cmd = subparser.add_parser("git-status",
        help="Runs git status on any git packages")
    git_status_cmd.set_defaults(func=CmdGitStatus())
    git_status_cmd.add_argument("-p", "-project-dir", dest="project_dir")
    
    git_update_cmd = subparser.add_parser("git-update")
    git_update_cmd.set_defaults(func=CmdGitUpdate())
    git_update_cmd.add_argument("-p", "-project-dir", dest="project_dir")
    
    snapshot_cmd = subparser.add_parser("snapshot",
        help="Creates a snapshot of required packages")
    snapshot_cmd.set_defaults(func=CmdSnapshot())
    snapshot_cmd.add_argument("-p", "-project-dir", dest="project_dir")
    snapshot_cmd.add_argument("-r", "--rls-deps", dest="rls", action="store_true",
        help="Uses release deps from project root instead of dev deps")
    snapshot_cmd.add_argument("snapshot_dir", 
            help="Specifies the directory where the snapshot will be created")

    if parser_ext is not None:
        for ext in parser_ext:
            ext(subparser)

    return parser

def main(project_dir=None):
    from .pkg_types.pkg_type_rgy import PkgTypeRgy

    # First things first: load any extensions
    import sys
    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points

    discovered_plugins = entry_points(group='ivpm.ext')
    parser_ext = []
    for p in discovered_plugins:
        try:
            mod = p.load()
            if hasattr(mod, "ivpm_subcommand"):
                parser_ext.append(getattr(mod, "ivpm_subcommand"))
            elif hasattr(mod, "ivpm_pkgtype"):
                pkg_types = []
                getattr(mod, "ivpm_pkgtype")(pkg_types)
                for pt in pkg_types:
                    PkgTypeRgy.inst().register(pt[0], pt[1], pt[2] if len(pt) > 2 else "")
        except Exception as e:
            print("Error: caught exception while loading IVPM extension %s (%s)" %(
                p.name,
                str(e)))
            raise e

    parser = get_parser(parser_ext)
    
    args = parser.parse_args()

    # If the user hasn't specified the project directory,
    # set the default
    if not hasattr(args, "project_dir") or getattr(args, "project_dir") is None:
        args.project_dir = project_dir

    args.func(args)

if __name__ == "__main__":
    main()
    




import os
import subprocess
from .proj_info import ProjInfo


def ivpm_popen(cmd, **kwargs):
    """
    Wrapper around subprocess.Popen that configures paths from 
    the nearest IVPM project.
    """
    ivpm_project = getattr(kwargs, "ivpm_project", None)

    if ivpm_project is None:
        # Search up from the invocation location
        cwd = os.getcwd()
        while cwd is not None and cwd != "/" and ivpm_project is None:
            if os.path.exists(os.path.join(cwd, "ivpm.yaml")):
                ivpm_project = cwd
            else:
                cwd = os.path.dirname(cwd)
    
    if ivpm_project is not None:
        # Update environment variables
        proj_info = ProjInfo.mkFromProj(ivpm_project)

        if proj_info is None:
            raise Exception("Failed to read ivpm.yaml @ %s" % ivpm_project)
        
        env = getattr(kwargs, "env", os.environ.copy())
        env["IVPM_PROJECT"] = ivpm_project
        env["IVPM_PACKAGES"] = os.path.join(ivpm_project, "packages")

        # Add the virtual-environment path
        if "PATH" in env.keys():
            env["PATH"] = os.path.join(ivpm_project, "packages/python/bin") + ":" + env["PATH"]
        else:
            env["PATH"] = os.path.join(ivpm_project, "packages/python/bin")

        for es in proj_info.env_settings:
            es.apply(env)

        kwargs["env"] = env

    return subprocess.Popen(cmd, **kwargs)



import os
import sys

class CmdShare(object):

    def __call__(self, args):
        ivpm_cmd_dir = os.path.dirname(os.path.abspath(__file__))
        ivpm_dir = os.path.dirname(ivpm_cmd_dir)
        ivpm_sharedir = os.path.join(ivpm_dir, "share")

        if args.path is not None:
            for p in args.path:
                ivpm_sharedir = os.path.join(ivpm_sharedir, p)
        sys.stdout.write(ivpm_sharedir.replace('\\', '/'))
#        print(ivpm_sharedir)



#****************************************************************************
#* package_handler_python.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import dataclasses as dc
import subprocess
import toposort
import os
from typing import Dict, List, Set
from ..project_ops_info import ProjectUpdateInfo
from ..utils import note, fatal, get_venv_python

from ..package import Package
from .package_handler import PackageHandler

@dc.dataclass
class PackageHandlerPython(PackageHandler):
    name="python"
    pkgs_info  : Dict[str,Package] = dc.field(default_factory=dict)
    src_pkg_s  : Set[str] = dc.field(default_factory=set)
    pypi_pkg_s : Set[str] = dc.field(default_factory=set)
    debug : bool = True

    def process_pkg(self, pkg: Package):
        print("process_pkg: %s (%s %s)" % (pkg.name, pkg.src_type, pkg.pkg_type))
        add = False
        if pkg.src_type == "pypi":
            self.pypi_pkg_s.add(pkg.name)
            add = True
        elif pkg.pkg_type is not None and pkg.pkg_type == PackageHandlerPython.name:
            self.src_pkg_s.add(pkg.name)
            add = True
        elif pkg.pkg_type is None and hasattr(pkg, "path"):
            # Check if there are known Python files
            print("Check files (%s)" % pkg.path)
            for py in ("setup.py", "setup.cfg", "pyproject.toml"):
                if os.path.isfile(os.path.join(pkg.path, py)):
                    print("Add to src_pkg_s")
                    add = True
                    self.src_pkg_s.add(pkg.name)
                    break            
        if add:
            pkg.pkg_type = PackageHandlerPython.name
            self.pkgs_info[pkg.name] = pkg
    
    def update(self, update_info : ProjectUpdateInfo):

        # Build up a dependency map for Python package installation        
        python_deps_m = {}
#        python_pkgs_s = set()

        # Collect the full set of packages
#        for pkg in self.packages:
#            python_pkgs_s.add(pkg.name)

        # Map between package name and a set of python
        # packages it depends on
        py_pkg_m = {}
        print("src_pkg_s: %s" % str(self.src_pkg_s))
        for pyp in self.src_pkg_s:
            print("pyp: %s" % pyp) 
            p = self.pkgs_info[pyp]
            if pyp not in python_deps_m.keys():
                python_deps_m[pyp] = set()

            if p.proj_info is not None:
                print("non-none proj_info")
                # TODO: see if the package specifies the package set
                if p.proj_info.has_dep_set(p.proj_info.target_dep_set):
                    for dp in p.proj_info.get_dep_set(p.proj_info.target_dep_set).keys():
                        if dp in self.pkgs_info.keys():
                            dp_p = self.pkgs_info[dp]
                            if dp_p.src_type != "pypi":
                                python_deps_m[pyp].add(dp)
                        # if dp in python_pkgs_s:
                        #     dp_p = pkgs_info[dp]
                        #     if dp_p.src_type != SourceType.PyPi:
                        #         python_deps_m[pyp].add(dp)
                else:
                    print("Warning: project %s does not contain its target dependency set (%s)" % (
                        p.proj_info.name,
                        p.proj_info.target_dep_set))
                    for d in p.proj_info.dep_set_m.keys():
                        print("Dep-Set: %s" % d)

        # Order the source packages based on their dependencies 
        it = toposort.toposort(python_deps_m)
        pysrc_pkg_order = list(it)
        if self.debug:
            print("python_deps_m: %s" % str(python_deps_m))
            print("pysrc_pkg_order: %s" % str(pysrc_pkg_order))

        python_deps_m = {}
        
#        # Collect all packages that have a setup dependency
#        for sdp in pkgs_info.setup_deps.keys():
#            print("Package %s has a setup dependency" % sdp)
#            if sdp not in python_deps_m.keys():
#                python_deps_m[sdp] = set()
#            for sdp_d in pkgs_info.setup_deps[sdp]:
#                python_deps_m[sdp].add(sdp_d)
#                setup_deps_s.add(sdp_d)
#
#        if self.debug:
#            print("python_deps_m: %s" % str(python_deps_m))
#        
#        pypkg_order = list(toposort(python_deps_m))
#        print("pypkg_order: %s" % str(pypkg_order))
#        
#        # Remove the original packages that asserted
#        # setup dependencies
#        for sdp in pkgs_info.setup_deps.keys():
#            if sdp not in setup_deps_s:
#                for pypkg_s in pypkg_order:
#                    pypkg_s.discard(sdp)
#        
#        if self.debug:           
#            print("ordered: %s" % str(pypkg_order))
#
        python_requirements_paths = []

        # Setup deps are a special category. We need to 
        # install them first -- possibly even before
        # installing other pypi packages
        setup_deps_s = set()
        # for pkg,deps in update_info.setup_deps.items():
        #     for dep in deps:
        #         if dep not in setup_deps_s:
        #             setup_deps_s.add(dep)
        #             if dep in self.pypi_pkg_s:
        #                 self.pypi_pkg_s.remove(dep)

        # for proj,deps in self.pkgs_info.setup_deps.items():
        #     for dep in deps:
        #         if dep not in setup_deps_s:
        #             setup_deps_s.add(dep)
        #             if dep in self.pypi_pkg_s:
        #                 self.pypi_pkg_s.remove(dep)
        print("setup_deps_s: %s" % str(setup_deps_s))

        if len(setup_deps_s) > 0:
            setup_deps_pkgs = []
            for dep in setup_deps_s:
                setup_deps_pkgs.append(self.pkgs_info[dep])

            requirements_path = os.path.join(
                update_info.deps_dir, "python_pkgs_%d.txt" % (
                len(python_requirements_paths)+1))
            self._write_requirements_txt(
                update_info.deps_dir,
                setup_deps_pkgs, 
                requirements_path)
            python_requirements_paths.append(requirements_path)

        # Next, create a requirements file for all
        # non-setup-dep PyPi packages
        python_pkgs = []
        print("pypi_pkg_s: %s" % str(self.pypi_pkg_s))
        for pypi_p in self.pypi_pkg_s:
            python_pkgs.append(self.pkgs_info[pypi_p])

        if len(python_pkgs) > 0:
            requirements_path = os.path.join(
                update_info.deps_dir, "python_pkgs_%d.txt" % (
                len(python_requirements_paths)+1))

            self._write_requirements_txt(
                update_info.deps_dir,
                python_pkgs, 
                requirements_path)
            python_requirements_paths.append(requirements_path)

        # Now, add requirement files for any source packages
        for pydep_s in pysrc_pkg_order:
            python_pkgs = []
            for key in pydep_s:
                
                # A future iteration does not need to install this
                self.pypi_pkg_s.discard(key)
                self.src_pkg_s.discard(key)
                
                # Note: for completeness, should collect Python 
                # packages known to be required by this pre-dep
                
                if key not in self.pkgs_info.keys():
                    raise Exception("Package %s not found in packages-info" % key)
                
                pkg : Package = self.pkgs_info[key]
                python_pkgs.append(pkg)
                
            if len(python_pkgs):
                requirements_path = os.path.join(
                    update_info.deps_dir, "python_pkgs_%d.txt" % (len(python_requirements_paths)+1))
                self._write_requirements_txt(
                    update_info.deps_dir,
                    python_pkgs, 
                    requirements_path)
                python_requirements_paths.append(requirements_path)
            
        if len(python_requirements_paths):
            import sys
            import platform
            ps = ";" if platform.system() == "Windows" else ":"
            env = os.environ.copy()
            env["PYTHONPATH"] = ps.join(sys.path)

            note("Installing Python dependencies in %d phases" % len(python_requirements_paths))
            for reqfile in python_requirements_paths:
                cwd = os.getcwd()
                os.chdir(os.path.join(update_info.deps_dir))
                cmd = [
                    get_venv_python(os.path.join(update_info.deps_dir, "python")),
                    "-m",
                    "ivpm.pywrap",
                    get_venv_python(os.path.join(update_info.deps_dir, "python")),
                    "-m",
                    "pip",
                    "install",
                    "-r",
                    reqfile]
            
                status = subprocess.run(cmd, env=env)
            
                if status.returncode != 0:
                    fatal("failed to install Python packages")
                os.chdir(cwd)        


    def _write_requirements_txt(self, 
                                packages_dir,
                                python_pkgs : List[Package],
                                file):
        """Writes a requirements file for pip to use in installing packages"""
        with open(file, "w") as fp:
            for pkg in python_pkgs:
                
                if hasattr(pkg, "url"):
                    # Editable package
                    # fp.write("-e file://%s/%s#egg=%s\n" % (
                    #     packages_dir.replace("\\","/"), 
                    #     pkg.name, 
                    #     pkg.name))
                    fp.write("-e %s/%s\n" % (
                        packages_dir.replace("\\","/"), 
                        pkg.name))
                else:
                    # PyPi package
                    if pkg.version is not None:
                        if pkg.version[0] in ['<','>','=']:
                            fp.write("%s%s\n" % (pkg.name, pkg.version))
                        else:
                            fp.write("%s==%s\n" % (pkg.name, pkg.version))
                    else:
                        fp.write("%s\n" % pkg.name)





#****************************************************************************
#* project_ops.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import dataclasses as dc
from .package import Package, SourceType
from .package_updater import PackageUpdater
from .handlers.package_handler_rgy import PackageHandlerRgy
from .project_ops_info import ProjectUpdateInfo
from .utils import fatal, note, get_venv_python, setup_venv

@dc.dataclass
class ProjectOps(object):
    root_dir : str
    debug : bool = False

    def update(self,
               dep_set : str = "default-dev",
               anonymous : bool = False,
               skip_venv : bool = False):
        from .proj_info import ProjInfo

        proj_info = ProjInfo.mkFromProj(self.root_dir)

        if proj_info is None:
            fatal("Failed to locate IVPM meta-data (eg ivpm.yaml)")
            
        deps_dir = os.path.join(self.root_dir, proj_info.deps_dir)
 
        # Ensure that we have a python virtual environment setup
        if not skip_venv:
            if not os.path.isdir(os.path.join(deps_dir, "python")):
                ivpm_python = setup_venv(os.path.join(deps_dir, "python"))
            else:
                note("python virtual environment already exists")
                ivpm_python = get_venv_python(os.path.join(deps_dir, "python"))
            
        print("********************************************************************")
        print("* Processing root package %s" % proj_info.name)
        print("********************************************************************")

        if self.debug:
            for self.dep_set in proj_info.dep_set_m.keys():
                print("DepSet: %s" % self.dep_set)
                for d in proj_info.dep_set_m[self.dep_set].packages.keys():
                    print("  Package: %s" % d)
                    
        if dep_set not in proj_info.dep_set_m.keys():
            raise Exception("Dep-set %s is not present" % dep_set)
        else:
            ds = proj_info.dep_set_m[dep_set]

        # If the root dependency set doesn't specify a source
        # for IVPM, auto-load it from PyPi
        if "ivpm" not in ds.packages.keys():
            print("Note: will install IVPM from PyPi")
            ivpm = Package("ivpm")
            ivpm.src_type = SourceType.PyPi
            ds.packages["ivpm"] = ivpm

        pkg_handler = PackageHandlerRgy.inst().mkHandler()
        updater = PackageUpdater(
            deps_dir, 
            pkg_handler,
            anonymous_git=anonymous)

        # Prevent an attempt to load the top-level project as a depedency
        updater.all_pkgs[proj_info.name] = None
        pkgs_info = updater.update(ds)

        print("Setup-deps: %s" % str(pkgs_info.setup_deps))

        # Finally, call the handlers to take care of project-level setup work
        update_info = ProjectUpdateInfo(deps_dir)
        pkg_handler.update(update_info)

    def status(self, dep_set : str = None):
        pass

    def sync(self, dep_set : str = None):
        pass





#****************************************************************************
#* package_http.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import httpx
import sys
import urllib
import dataclasses as dc
from .package_file import PackageFile
from ..project_ops_info import ProjectUpdateInfo
from ..utils import note
from ..package import SourceType2Ext

class PackageHttp(PackageFile):

    def update(self, update_info : ProjectUpdateInfo):
        pkg_dir = os.path.join(update_info.deps_dir, self.name)
        self.path = pkg_dir.replace("\\", "/")

        if os.path.isdir(pkg_dir):
            note("Skipping %s, since it is already loaded" % self.name)
        else:
            # Need to fetch, then unpack these
            download_dir = os.path.join(update_info.deps_dir, ".download")
                
            if not os.path.isdir(download_dir):
                os.makedirs(download_dir)

            if self.unpack:
                pkg_path = os.path.join(download_dir, 
                                        os.path.basename(self.url))
            else:
                pkg_path = os.path.join(update_info.deps_dir, self.name)
                    
            # TODO: should this be an option?   
            remove_pkg_src = True

            print("http: unpack=%s url=%s" % (str(self.unpack), str(self.url)))
            r = httpx.get(self.url, follow_redirects=True)
            with open(pkg_path, "wb") as f:
                f.write(r.content)

            if self.unpack:
                self._install(pkg_path, pkg_dir)
                os.unlink(os.path.join(download_dir, 
                                       os.path.basename(self.url)))
            else:
                # 
                pass
            
    @staticmethod
    def create(name, opts, si) -> 'PackageHttp':
        pkg = PackageHttp(name)
        pkg.process_options(opts, si)
        return pkg








'''
Created on Jun 8, 2021

@author: mballance
'''
import os

from ivpm.packages_info import PackagesInfo
from ivpm.project_ops import ProjectOps
from ivpm.out_wrapper import OutWrapper
from ivpm.package_updater import PackageUpdater
from ivpm.package import Package, PackageType, SourceType
from typing import List


class CmdUpdate(object):
    
    def __init__(self):
        self.debug = False
        pass
    
    def __call__(self, args):
        if args.project_dir is None:
            # If a default is not provided, use the current directory
            print("Note: project_dir not specified ; using working directory")
            args.project_dir = os.getcwd()

        ds_name = "default-dev"

        if hasattr(args, "dep_set") and args.dep_set is not None:
            ds_name = args.dep_set

        ProjectOps(args.project_dir).update(
            dep_set=ds_name,
            anonymous=args.anonymous)



   
    



#****************************************************************************
#* project_opts_info.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import dataclasses as dc
import enum

class FileStatus(enum.Enum):
    Unknown = "U"
    Modified = "M"
    Added = "A"
    Deleted = "D"

@dc.dataclass
class ProjectOpsInfo(object):
    deps_dir : str

@dc.dataclass
class ProjectSyncInfo(ProjectOpsInfo):
    pass

@dc.dataclass
class ProjectStatusInfo(ProjectOpsInfo):
    pass

@dc.dataclass
class ProjectStatusResult(object):
    package : str
    path : str
    status : FileStatus

@dc.dataclass
class ProjectUpdateInfo(ProjectOpsInfo):
    anonymous_git : bool = False





#!/usr/bin/env python3

#****************************************************************************
#* ivpm.py
#* 
#* This is the bootstrap ivpm.py script that is included with each project.
#* This script ensures that the *actual* ivpm is downloaded in the 
#* project packages dir
#****************************************************************************
import os.path
import sys
import subprocess

#********************************************************************
#* download_ivpm
#*
#* 
#********************************************************************
def download_ivpm(packages_dir):
    if os.path.isdir(packages_dir) == False:
        os.makedirs(packages_dir)
       
    cwd = os.getcwd()
    os.chdir(packages_dir)
    status = os.system("git clone https://github.com/mballance/ivpm.git")
    os.chdir(cwd);

def main():
    scripts_dir = os.path.dirname(os.path.realpath(__file__))
    project_dir = os.path.dirname(scripts_dir)
    packages_dir = project_dir + "/packages"
    ivpm_dir = packages_dir + "/ivpm"
    
    help_requested = False;
    
    for arg in sys.argv:
        if (arg == "-help" or arg == "--help" 
            or arg == "-h" or arg == "--h" 
            or arg == "-?"):
            help_requested = True
            
    
    # First see if we've already downloaded IVPM
    if os.path.isdir(ivpm_dir) == False:
        if (help_requested):
            print("Local help");
            sys.exit(1);
        download_ivpm(packages_dir)
        
    # Bring in the actual IVPM script and call it
    sys.path.insert(0, ivpm_dir + "/src")
    from ivpm.__main__ import main
    main(project_dir)
    
if __name__ == "__main__":
    main()



#****************************************************************************
#* packages_info.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on: Jan 19, 2020
#*     Author: mballance
#*
#****************************************************************************
from typing import Dict, List, Set
from ivpm.package import Package

class PackagesInfo():
    """
    Holds information about a set of packages. Holds data
    from one dep-set in an IVPM file.
    """
    
    def __init__(self, name):
        self.name = name
        self.packages : Dict[str,Package] = {}
        
        # Map of package name to set of packages
        # required for setup. This is Python-specific
        self.setup_deps : Dict[str, Set[str]] = {}
        self.options = {}

    def keys(self):
        return self.packages.keys()
    
    def add_package(self, pkg : Package):
        self.packages[pkg.name] = pkg

    def get_options(self, package):
        if package in self.options.keys():
            return self.options[package]
        else:
            return {}

    def set_options(self, package, options):
        self.options[package] = options

    def pop(self, key):
        self.packages.pop(key)

    def __getitem__(self, key):
        return self.packages[key]

    def __setitem__(self, key, value):
        self.packages[key] = value
        
    def copy(self) -> 'PackagesInfo':
        ret = PackagesInfo(self.name)
        ret.packages = self.packages.copy()
        ret.options  = self.options.copy()
        
        return ret



import os
import shutil
import sys
import subprocess
import unittest

sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(
        os.path.dirname(os.path.abspath(__file__)))), "src"));

from ivpm.project_ops import ProjectOps

class TestBase(unittest.TestCase):

    def setUp(self) -> None:
        unit_dir = os.path.dirname(os.path.abspath(__file__))
        self.data_dir = os.path.join(unit_dir, "data")
        os.environ["DATA_DIR"] = self.data_dir
        test_dir = os.path.dirname(unit_dir)

        rundir = os.path.join(test_dir, "rundir")
        self.testdir = os.path.join(rundir, "test")

        if os.path.isdir(self.testdir):
            shutil.rmtree(self.testdir)
        os.makedirs(self.testdir)

        return super().setUp()
    
    def tearDown(self) -> None:
        return super().tearDown()
    
    def mkFile(self, filename, content):
        fullpath = os.path.join(self.testdir, filename)
        parent = os.path.dirname(fullpath)

        if not os.path.isdir(parent):
            os.makedirs(parent)
        
        with open(fullpath, "w") as fp:
            fp.write(content)

    def ivpm_update(self, dep_set="default-dev", anonymous=False, skip_venv=False):
        
        ProjectOps(self.testdir).update(
                dep_set=dep_set, 
                anonymous=anonymous,
                skip_venv=skip_venv
            )
    
    def ivpm_sync(self, dep_set=None):
        ProjectOps(self.testdir).sync(dep_set=dep_set)
        
    def exec(self, cmd, cwd=None):
        return subprocess.check_output(cmd, cwd=cwd).decode("utf-8")





#****************************************************************************
#* project_updater.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on: Jun 22, 2021
#*     Author: mballance
#*
#****************************************************************************
import os
import shutil
import subprocess
import sys
import tarfile
import urllib
from zipfile import ZipFile

from ivpm.msg import note, fatal, warning
from ivpm.package import Package, SourceType, SourceType2Ext, PackageType
from ivpm.packages_info import PackagesInfo
from ivpm.proj_info import ProjInfo
from typing import Dict
from ivpm.utils import get_venv_python
from .project_ops_info import ProjectUpdateInfo


class PackageUpdater(object):
    
    def __init__(self, 
                 deps_dir, 
                 pkg_handler,
                 anonymous_git=False,
                 load=True):
        self.debug = False
        self.deps_dir = deps_dir
        self.pkg_handler = pkg_handler
        self.all_pkgs = PackagesInfo("root")
        self.new_deps = []
        self.anonymous_git = anonymous_git
        self.load = load
        pass
    
    def update(self, pkgs : PackagesInfo) -> PackagesInfo:
        """
        Updates the specified packages, handling dependencies
        The 'pkgs' parameter holds the dependency information
        from the root project
        """
        count = 1

        pkg_q = []
        
        if len(pkgs.keys()) == 0:
            print("No packages")
        
        for key in pkgs.keys():
            print("Package: %s" % key)
            pkg_q.append(pkgs[key])

        if not os.path.isdir(self.deps_dir):
            os.makedirs(self.deps_dir)
            
        while True:        
            pkg_deps = {}
            
            # Process this batch of packages
            while len(pkg_q) > 0:
                pkg : Package = pkg_q.pop(0)
                
                self.all_pkgs[pkg.name] = pkg
                
                proj_info : ProjInfo = self._update_pkg(pkg)

                # proj_info contains info on any setup-deps that
                # might be required
                if proj_info is not None:
                    for sd in proj_info.setup_deps:
                        print("Add setup-dep %s to package %s" % (sd, pkg.name))
                        if pkg.name not in self.all_pkgs.setup_deps.keys():
                            self.all_pkgs.setup_deps[pkg.name] = set()
                        self.all_pkgs.setup_deps[pkg.name].add(sd)

                    if proj_info.process_deps:
                        if not proj_info.has_dep_set(pkg.dep_set):
                            warning("package %s does not contain specified dep-set %s ; skipping" % (proj_info.name, pkg.dep_set))
                            continue
                        else:
                            note("Loading package %s dependencies from dep-set %s" % (proj_info.name, pkg.dep_set))

                        note("Processing dep-set %s of project %s" % (
                            pkg.dep_set,
                            pkg.name))                        

                        ds : PackagesInfo = proj_info.get_dep_set(pkg.dep_set)
                        for d in ds.packages.keys():
                            dep = ds.packages[d]
                    
                            if dep.name not in pkg_deps.keys():
                                pkg_deps[dep.name] = dep
                            else:
                                # TODO: warn about possible version conflict?
                                pass
            
            # Collect new dependencies and add to queue
            for key in pkg_deps.keys():
                if not key in self.all_pkgs.keys():
                    # New package
                    pkg_q.append(pkg_deps[key])
            note("%d new dependencies from iteration %d" % (len(pkg_q), count))
                    
            if len(pkg_q) == 0:
                # We're done
                break
            
            count += 1
            
        return self.all_pkgs
    
    def _update_pkg(self, pkg : Package) -> ProjInfo:
        """Loads a single package. Returns any dependencies"""
        must_update=False

        update_info = ProjectUpdateInfo(self.deps_dir)
  
        print("********************************************************************")
        print("* Processing package %s" % pkg.name)
        print("********************************************************************")


        pkg_dir = os.path.join(self.deps_dir, pkg.name)
        pkg.path = pkg_dir.replace("\\", "/")

        pkg.proj_info = pkg.update(update_info)

        # Notify the package handlers after the source is 
        # loaded so they can take further action if required 
        self.pkg_handler.process_pkg(pkg)
        
        # Ensure that we use the requested dep-set
        if pkg.proj_info is not None:
            pkg.proj_info.target_dep_set = pkg.dep_set
            pkg.proj_info.process_deps = pkg.process_deps
        
        return pkg.proj_info

    

    




#****************************************************************************
#* cmd_pkg_flags.py
#*
#* Copyright 2022 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************

from ..pkg_info.pkg_info_rgy import PkgInfoRgy
from ..pkg_info.pkg_info_loader import PkgInfoLoader
from ..pkg_info.pkg_compile_flags import PkgCompileFlags

class CmdPkgInfo(object):

    def __init__(self):
        pass

    def __call__(self, args):
        rgy = PkgInfoRgy.inst()
        pkg_names = args.pkgs if isinstance(args.pkgs, list) else [args.pkgs]
        pkgs = []

        for pn in pkg_names:
            if not rgy.hasPkg(pn):
                raise Exception("Failed to find package %s" % pn)
            pkgs.append(rgy.getPkg(pn))

        if args.type == "flags":
            flags = PkgCompileFlags().flags(pkgs)
            print("%s" % " ".join(flags))
        elif args.type == "paths":
            paths = PkgCompileFlags().paths(pkgs, args.kind)
            print("%s" % " ".join(paths))
        elif args.type == "libdirs":
            paths = PkgCompileFlags().ldirs(pkgs, args.kind)
            print("%s" % " ".join(paths))
        elif args.type == "libs":
            paths = PkgCompileFlags().libs(pkgs, args.kind)
            print("%s" % " ".join(paths))
        else:
            raise Exception("Unimplemented pkg-info kind %s" % args.type)





#****************************************************************************
#* package_git.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import sys
import subprocess
import dataclasses as dc
from .package_url import PackageURL
from ..proj_info import ProjInfo
from ..project_ops_info import ProjectUpdateInfo, ProjectStatusInfo, ProjectSyncInfo
from ..utils import note, fatal

@dc.dataclass
class PackageGit(PackageURL):
    branch : str = None
    commit : str = None
    tag : str = None
    depth : str = None
    anonymous : bool = None

    def update(self, update_info : ProjectUpdateInfo) -> ProjInfo:
        pkg_dir = os.path.join(update_info.deps_dir, self.name)
        self.path = pkg_dir.replace("\\", "/")

        if os.path.exists(pkg_dir):
            note("package %s is already loaded" % self.name)
        else:
            note("loading package %s" % self.name)

            cwd = os.getcwd()
            os.chdir(update_info.deps_dir)
            sys.stdout.flush()

            git_cmd = ["git", "clone"]
        
            if self.depth is not None:
                git_cmd.extend(["--depth", str(self.depth)])

            if self.branch is not None:
                git_cmd.extend(["-b", str(self.branch)])

            # Modify the URL to use SSH/key-based clones
            # unless anonymous cloning was requested
            use_anonymous = update_info.anonymous_git

            if self.anonymous is not None:
                use_anonymous = self.anonymous

            if not use_anonymous:
                print("NOTE: using dev URL")
                delim_idx = self.url.find("://")
                url = self.url[delim_idx+3:]
                first_sl_idx = url.find('/')
                url = "git@" + url[:first_sl_idx] + ":" + url[first_sl_idx+1:]
                print("Final URL: %s" % url)
                git_cmd.append(url)
            else:
                print("NOTE: using anonymous URL")
                git_cmd.append(self.url)

            # Clone to a directory with same name as package        
            git_cmd.append(self.name)
            
            print("git_cmd: \"" + str(git_cmd) + "\"")
            status = subprocess.run(git_cmd)
            os.chdir(cwd)
        
            if status.returncode != 0:
                fatal("Git command \"%s\" failed" % str(git_cmd))

            # Checkout a specific commit            
            if self.commit is not None:
                os.chdir(os.path.join(self.deps_dir, self.name))
                git_cmd = "git reset --hard %s" % self.commit
                status = os.system(git_cmd)
            
                if status != 0:
                    fatal("Git command \"%s\" failed" % str(git_cmd))
                os.chdir(cwd)
            
        
            # TODO: Existence of .gitmodules should trigger this
            if os.path.isfile(os.path.join(update_info.deps_dir, self.name, ".gitmodules")):
                os.chdir(os.path.join(update_info.deps_dir, self.name))
                sys.stdout.flush()
                status = os.system("git submodule update --init --recursive")
                os.chdir(cwd)

        return ProjInfo.mkFromProj(pkg_dir)
    
    def status(self, status_info : ProjectStatusInfo):
        pass
    
    def sync(self, sync_info : ProjectSyncInfo):
        if not os.path.isdir(os.path.join(sync_info.deps_dir, dir, ".git")):
            fatal("Package \"" + dir + "\" is not a Git repository")
        print("Package: " + dir)
        try:
            branch = subprocess.check_output(
                ["git", "branch"],
                cwd=os.path.join(sync_info.deps_dir, dir))
        except Exception as e:
            fatal("failed to get branch of package \"" + dir + "\"")

        branch = branch.strip()
        if len(branch) == 0:
            fatal("branch is empty")

        branch_lines = branch.decode().splitlines()
        branch = None
        for bl in branch_lines:
            if bl[0] == "*":
                branch = bl[1:].strip()
                break
        if branch is None:
            fatal("Failed to identify branch")

        status = subprocess.run(
            ["git", "fetch"],
            cwd=os.path.join(sync_info.deps_dir, dir))
        if status.returncode != 0:
            fatal("Failed to run git fetch on package %s" % dir)
        status = subprocess.run(
            ["git", "merge", "origin/" + branch],
            cwd=os.path.join(sync_info.deps_dir, dir))
        if status.returncode != 0:
            fatal("Failed to run git merge origin/%s on package %s" % (branch, dir))
    
    def process_options(self, opts, si):
        super().process_options(opts, si)

        if "anonymous" in opts.keys():
            self.anonymous = opts["anonymous"]
                
        if "depth" in opts.keys():
            self.depth = opts["depth"]
                
        if "dep-set" in opts.keys():
            self.dep_set = opts["dep-set"]
               
        if "branch" in opts.keys():
            self.branch = opts["branch"]
                
        if "commit" in opts.keys():
            self.commit = opts["commit"]
               
        if "tag" in opts.keys():
            self.tag = opts["tag"]


    @staticmethod
    def create(name, opts, si) -> 'PackageGit':
        pkg = PackageGit(name)
        pkg.process_options(opts, si)
        return pkg




'''
Created on Jun 8, 2021

@author: mballance
'''
import os
import subprocess
import sys

from ivpm.arg_utils import ensure_have_project_dir
from ivpm.msg import fatal
from ..project_ops import ProjectOps


class CmdSync(object):
    
    def __init__(self):
        pass
    
    def __call__(self, args):
        if args.project_dir is None:
            args.project_dir = os.getcwd()
    
        packages_dir = os.path.join(args.project_dir, "packages")
    
        # After that check, go ahead and just check directories
        for dir in os.listdir(packages_dir):
            if os.path.isdir(os.path.join(packages_dir, dir, ".git")):
                print("Package: " + dir)
                cwd = os.getcwd()
                os.chdir(packages_dir + "/" + dir)
                try:
                    branch = subprocess.check_output(["git", "branch"])
                except Exception as e:
                    print("Note: Failed to get branch of package \"" + dir + "\"")
                    continue

                branch = branch.strip()
                if len(branch) == 0:
                    raise Exception("Error: branch is empty")

                branch_lines = branch.decode().splitlines()
                branch = None
                for bl in branch_lines:
                    if bl[0] == "*":
                        branch = bl[1:].strip()
                        break
                if branch is None:
                    raise Exception("Failed to identify branch")

                status = subprocess.run(["git", "fetch"])
                if status.returncode != 0:
                    fatal("Failed to run git fetch on package %s" % dir)
                status = subprocess.run(["git", "merge", "origin/" + branch])
                if status.returncode != 0:
                    fatal("Failed to run git merge origin/%s on package %s" % (branch, dir))
                os.chdir(cwd)
            elif os.path.isdir(packages_dir + "/" + dir):
                print("Note: skipping non-Git package \"" + dir + "\"")
                sys.stdout.flush()        
        pass



#****************************************************************************
#* pkg_info_loader.py
#*
#* Copyright 2022 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import importlib
import typing
from .pkg_info import PkgInfo

class PkgInfoLoader(object):

    def __init__(self):
        self._pkg_info_m = {}
        pass

    def load(self, root_pkg) -> PkgInfo:
        self._pkg_info_m.clear()

        pi = self._load_pkg(root_pkg)

        return pi
    
    def load_pkgs(self, root_pkgs) -> typing.List[PkgInfo]:
        ret = []
        self._pkg_info_m.clear()

        for p in root_pkgs:
            if p in self._pkg_info_m.keys():
                ret.append(self._pkg_info_m[p])
            else:
                ret.append(self._load_pkg(p))

        return ret


    def _load_pkg(self, name):
        pkg = importlib.import_module(name)

        if hasattr(pkg, "pkginfo"):
            # Load the dedicated class
            pkginfo_m = getattr(pkg, "pkginfo")
            if not hasattr(pkginfo_m, "PkgInfo"):
                raise Exception("Failed to find pkginfo:PkgInfo in %s" % name)
            
            pkg_info = getattr(pkginfo_m, "PkgInfo")()

            for dep in pkg_info.getDeps():
                pass
        else:
            pkg_info = PkgInfo(name)

            deps = []
            if hasattr(pkg, "get_deps"):
                deps.extend(pkg.get_deps())

            if hasattr(pkg, "get_libs"):
                pkg_info._libs.extend(pkg.get_libs())

            if hasattr(pkg, "get_libdirs"):
                pkg_info._libdirs.extend(pkg.get_libdirs())

            if hasattr(pkg, "get_incdirs"):
                pkg_info._incdirs.extend(pkg.get_incdirs())

            self._pkg_info_m[name] = pkg_info

            for d in deps:
                if d not in self._pkg_info_m.keys():
                    dep = self._load_pkg(d)
                    self._pkg_info_m[d] = dep
                    pkg_info._deps.append(dep)
                else:
                    pkg_info._deps.append(self._pkg_info_m[d])

        return pkg_info







import os
import sys
import subprocess
from ivpm.packages_info import PackagesInfo
from ivpm.proj_info import ProjInfo
from ivpm.utils import get_venv_python
from ivpm.msg import note, fatal, warning
from ivpm.package_updater import PackageUpdater
from ivpm.package import Package, PackageType, SourceType
from toposort import toposort
from typing import List

class CmdBuild(object):

    def __init__(self):
        self.debug = False
        pass

    def __call__(self, args):
        if args.project_dir is None:
            # If a default is not provided, use the current directory
            print("Note: project_dir not specified ; using working directory")
            args.project_dir = os.getcwd()
            
        proj_info = ProjInfo.mkFromProj(args.project_dir)

        if proj_info is None:
            fatal("Failed to locate IVPM meta-data (eg ivpm.yaml)")
            
        packages_dir = os.path.join(args.project_dir, "packages")
 
        # Ensure that we have a python virtual environment setup
        if os.path.isdir(os.path.join(packages_dir, "python")):
            ivpm_python = get_venv_python(os.path.join(packages_dir, "python"))
        else:
            raise Exception("packages/python does not exist ; ivpm update must be run before ivpm build")
            
        print("********************************************************************")
        print("* Processing root package %s" % proj_info.name)
        print("********************************************************************")

        if self.debug:
            for ds_name in proj_info.dep_set_m.keys():
                print("DepSet: %s" % ds_name)
                for d in proj_info.dep_set_m[ds_name].packages.keys():
                    print("  Package: %s" % d)
                    
        ds_name = "default-dev"
        
        if getattr(args, "dep_set") is not None:
            ds_name = args.dep_set
            
        if ds_name not in proj_info.dep_set_m.keys():
            raise Exception("Dep-set %s is not present" % ds_name)
        else:
            ds = proj_info.dep_set_m[ds_name]

        # TODO: need a way to step through packages without updating
        updater = PackageUpdater(packages_dir, load=False)
        # Prevent an attempt to load the top-level project as a depedency
        updater.all_pkgs[proj_info.name] = None
        pkgs_info : PackagesInfo = updater.update(ds)

        # Remove the root package before processing what's left
        pkgs_info.pop(proj_info.name)

        # Build up a dependency map for Python package installation        
        python_deps_m = {}
        python_pkgs_s = set()

        # Collect the full set of packages
        for pid in pkgs_info.keys():
            p = pkgs_info[pid]
            if p.pkg_type == PackageType.Python and p.src_type != SourceType.PyPi:
                python_pkgs_s.add(pid)

        pypi_pkg_s = set() # List of packages on PyPi to install
        # Map between package name and a set of python
        # packages it depends on
        py_pkg_m = {}
        for pyp in python_pkgs_s:
            p = pkgs_info[pyp]
            if p.src_type == SourceType.PyPi:
                pypi_pkg_s.add(p.name)
            else:
                # Read the package-info
                if pyp not in python_deps_m.keys():
                    python_deps_m[pyp] = set()

                proj_info = ProjInfo.mkFromProj(p.path)

                if proj_info is None:
                    continue

                # TODO: see if the package specifies the package set
                if proj_info.has_dep_set(proj_info.target_dep_set):
                    for dp in proj_info.get_dep_set(proj_info.target_dep_set).keys():
                        if dp in python_pkgs_s:
                            dp_p = pkgs_info[dp]
                            if dp_p.src_type != SourceType.PyPi:
                                python_deps_m[pyp].add(dp)
                else:
                    print("Warning: project %s does not contain its target dependency set (%s)" % (
                        proj_info.name,
                        proj_info.target_dep_set))
                    for d in proj_info.dep_set_m.keys():
                        print("Dep-Set: %s" % d)

        # Order the source packages based on their dependencies 
        pysrc_pkg_order = list(toposort(python_deps_m))
        if self.debug:
            print("python_deps_m: %s" % str(python_deps_m))
            print("pysrc_pkg_order: %s" % str(pysrc_pkg_order))

        env = os.environ.copy()
        env["DEBUG"] = "1" if args.debug else "0"
        for pkgs in pysrc_pkg_order:
            for pkg in pkgs:
                if os.path.isfile(os.path.join(packages_dir, pkg, "setup.py")):
                    cmd = [
                        sys.executable,
                        'setup.py',
                        'build_ext',
                        '--inplace'
                    ]
                    result = subprocess.run(
                        cmd,
                        env=env,
                        cwd=os.path.join(packages_dir, pkg))

                    if result.returncode != 0:
                        raise Exception("Failed to build package %s" % pkg)
                else:
                    note("Skipping Python project without setup.py")



